
from abc import ABC, abstractmethod

from gym.spaces import Discrete
import numpy as np

from .policy import Policy


class QTablePolicy(Policy, ABC):
    """
    A policy that explicitly stores and updates a Q-table.

    This requires Discrete observation space and action space.
    """
    def __init__(self, q_table=None, **kwargs):
        super().__init__(**kwargs)
        if q_table is None:
            self._q_table = np.random.normal(0, 1, (self.observation_space.n, self.action_space.n))
        else:
            self.q_table = q_table

    @property
    def q_table(self):
        """
        The Q-table stores the value of each state-action pair.

        The rows of the table should be the observations and the columns should
        be actions.
        """
        return self._q_table

    @q_table.setter
    def q_table(self, value):
        assert isinstance(value, np.ndarray), "Q-Table must be a numpy array."
        assert len(value.shape) == 2, "Q-Table must have two dimensions."
        assert value.shape[0] == self.observation_space.n, \
            "The number of rows in the q table must be the same as the size of " + \
            "the observation space."
        assert value.shape[1] == self.action_space.n, \
            "The number of columns in the q table must be the same as the size of " + \
            "the action space."
        self._q_table = value

    @property
    def action_space(self):
        """
        The space of possible output actions for this policy.

        Must be Discrete.
        """
        return self._action_space

    @action_space.setter
    def action_space(self, value):
        assert isinstance(value, Discrete), "Action space must be Discrete."
        self._action_space = value

    @property
    def observation_space(self):
        """
        The space of possible input observations for this policy.

        Must be Discrete.
        """
        return self._observation_space

    @observation_space.setter
    def observation_space(self, value):
        assert isinstance(value, Discrete), "Observation space must be Discrete."
        self._observation_space = value

    @abstractmethod
    def probability(self, obs, action, **kwargs):
        """
        Calculate the probability of choosing this action given this observation.
        """
        pass

    def update(self, obs, action, value, **kwargs):
        """
        Set the q_table at (obs, action) to the value.
        """
        self.q_table[obs, action] = value


class EpsilonSoftPolicy(QTablePolicy):
    """
    Choose random action with some probability epsilon, otherwise choose optimal action.

    The EpsilonSoftPolicy will sample a uniform distribution between 0 and 1. If the sampled
    value is less than epsilon, then the policy will randomly choose an action. Otherwise, it
    will return the optimal action.
    """
    def __init__(self, epsilon=0.1, **kwargs):
        super().__init__(**kwargs)
        self.epsilon = epsilon

    @property
    def epsilon(self):
        """
        The probability of choosing a random action.
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        assert 0 <= value <= 1.0, "Epsilon must be between 0 and 1."
        self._epsilon = value

    def compute_action(self, obs, **kwargs):
        if np.random.uniform(0, 1) < self.epsilon:
            return np.random.randint(0, self.q_table[obs].size)
        else:
            return np.argmax(self.q_table[obs])

    def probability(self, obs, action, **kwargs):
        if action == np.argmax(self.q_table[obs]): # Optimal action
            return 1 - self.epsilon + self.epsilon / self.q_table[obs].size
        else: # Nonoptimal action
            return self.epsilon / self.q_table[obs].size



from gym.spaces import Space, Discrete, MultiBinary, MultiDiscrete, Box, Dict, Tuple


def check_space(space, strict=False):
    """
    Ensure that the space is a gym Space, including all nested spaces.

    strict (bool), default False:
        If strict is True, then the recursion rule is that every subspace must be
        a gym space. If strict is False, then the recursion rule is that every subspace
        must be a gym space OR a dict or tuple. In this way, we allow the space
        to be iteratively built and assume that the final wrapping to Dict or Tuple
        has yet to occur.
    """
    if isinstance(space, (Discrete, MultiDiscrete, MultiBinary, Box)):
        return True
    elif isinstance(space, Dict):
        return all([check_space(sub_space) for sub_space in space.spaces.values()])
    elif isinstance(space, Tuple):
        return all([check_space(sub_space) for sub_space in space.spaces])
    elif not strict:
        if isinstance(space, dict):
            return all([check_space(sub_space) for sub_space in space.values()])
        elif isinstance(space, tuple):
            return all([check_space(sub_space) for sub_space in space])
    else:
        return False


def make_dict(space):
    """
    Convert a hierarchical space into a gym space by recursively moving through
    the layers and converting the subspaces to gym spaces. Unsafe, modifies the
    items of the input as it moves through them.
    """
    assert isinstance(space, (dict, Space)), "Cannot convert this to a Dict."
    for key, subspace in space.items():
        if isinstance(subspace, dict):
            space[key] = make_dict(subspace)
        else:
            assert isinstance(subspace, Space), "Cannot convert this to a Dict."

    return Dict(space) if type(space) is dict else space



from abc import ABC, abstractmethod

from abmarl.sim import AgentBasedSimulation


class SimulationManager(ABC):
    """
    Control interaction between Trainer and AgentBasedSimulation.

    A Manager implmenents the reset and step API, by which it calls the AgentBasedSimulation
    API, using the getters within reset and step to accomplish the desired control
    flow.

    Attributes:
        sim: The AgentBasedSimulation.
        agents: The agents that are in the AgentBasedSimulation.
    """
    def __init__(self, sim):
        assert isinstance(sim, AgentBasedSimulation), \
            "SimulationManager can only interface with AgentBasedSimulation."
        self.sim = sim
        self.agents = sim.agents

    @abstractmethod
    def reset(self, **kwargs):
        """
        Reset the simulation.

        Returns:
            The first observation of the agent(s).
        """
        pass

    @abstractmethod
    def step(self, action_dict, **kwargs):
        """
        Step the simulation forward one discrete time-step.

        Args:
            action_dict:
                Dictionary mapping agent(s) to their actions in this time step.

        Returns:
            The observations, rewards, done status, and info for the agent(s) whose
            actions we expect to receive next.

            Note: We do not necessarily return anything for the agent whose actions
            we just received in this time-step. This behavior is defined by each
            Manager.
        """
        pass

    def render(self, **kwargs):
        self.sim.render(**kwargs)



from abmarl.tools import utils as adu


def run(full_config_path, parameters):
    """Debug the SimulationManagers from the config_file."""

    # Load the experiment as a module
    experiment_mod = adu.custom_import_module(full_config_path)
    title = "DEBUG_" + experiment_mod.params['experiment']['title']

    # Copy the configuration module to the output directory
    import os
    import shutil
    import time
    base = experiment_mod.params['ray_tune'].get('local_dir', os.path.expanduser("~"))
    output_dir = os.path.join(
        base, 'abmarl_results/{}_{}'.format(
            title, time.strftime('%Y-%m-%d_%H-%M')
        )
    )
    experiment_mod.params['ray_tune']['local_dir'] = output_dir
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copy(full_config_path, output_dir)

    # Simulation loop
    from pprint import pprint
    if parameters.render:
        from matplotlib import pyplot as plt
    sim = experiment_mod.params['experiment']['sim_creator'](
        experiment_mod.params['ray_tune']['config']['env_config']
    )
    agents = sim.unwrapped.agents
    for i in range(parameters.episodes):
        # Setup dump files
        with open(os.path.join(output_dir, f"Episode_{i}.txt"), 'w') as debug_dump:
            if parameters.render:
                fig = plt.figure()
            obs = sim.reset()
            done = {agent: False for agent in obs}
            if parameters.render:
                sim.render(fig=fig)
                plt.pause(1e-16)
            debug_dump.write("Reset:\n")
            pprint(obs, stream=debug_dump)
            for j in range(parameters.steps_per_episode): # Data generation
                action = {
                    agent_id: agents[agent_id].action_space.sample()
                    for agent_id in obs if not done[agent_id]
                }
                obs, reward, done, info = sim.step(action)
                if parameters.render:
                    sim.render(fig=fig)
                    plt.pause(1e-16)
                debug_dump.write(f"\nStep {j}:\n")
                pprint(action, stream=debug_dump)
                pprint(obs, stream=debug_dump)
                pprint(reward, stream=debug_dump)
                pprint(done, stream=debug_dump)
                if done['__all__']:
                    break
            if parameters.render:
                plt.close(fig)




import numpy as np

from abmarl.algs.monte_carlo import off_policy
from abmarl.sim.corridor import MultiCorridor as Corridor
from abmarl.managers import AllStepManager
from abmarl.sim.wrappers import RavelDiscreteWrapper
from abmarl.external import GymWrapper
from abmarl.policies.q_table_policy import EpsilonSoftPolicy


def test_off_policy():
    np.random.seed(24)
    sim = AllStepManager(RavelDiscreteWrapper(Corridor(num_agents=1)))
    sim, q_table, policy = off_policy(sim, iteration=1000, horizon=10)

    assert isinstance(sim, GymWrapper)
    assert isinstance(sim.sim, AllStepManager)
    assert isinstance(sim.sim.sim, RavelDiscreteWrapper)
    assert isinstance(sim.sim.sim.sim, Corridor)

    assert q_table.shape == (sim.observation_space.n, sim.action_space.n)
    assert isinstance(policy, EpsilonSoftPolicy)

    policy.epsilon = 0
    obs = sim.reset()
    for _ in range(10):
        action = policy.compute_action(obs)
        obs, reward, done, info = sim.step(action)
        if done:
            break
    assert done




import numpy as np

from abmarl.sim.gridworld.grid import Grid
from abmarl.sim.gridworld.actor import MovingAgent

grid = Grid(5,6)

moving_agents = {
    'agent0': MovingAgent(
        id='agent0', initial_position=np.array([3, 4]), encoding=1, move_range=1
    ),
    'agent1': MovingAgent(
        id='agent1', initial_position=np.array([2, 2]), encoding=2, move_range=2
    ),
    'agent2': MovingAgent(
        id='agent2', initial_position=np.array([0, 1]), encoding=1, move_range=1
    ),
    'agent3': MovingAgent(
        id='agent3', initial_position=np.array([3, 1]), encoding=3, move_range=3
    ),
}




import numpy as np


def create_grid_and_mask(agent, grid, mask_range, agents):
    """
    Generate a local grid and a mask.

    Create a local grid centered around the agents location and fill it will the
    values from the grid.

    Blocking agents can mask the grid from other agent's, restricting their
    ability to observe, attack, move, etc. We calculate the masking by drawing
    rays from the center of the agent's position to the edges of the other agents'
    cell. All cells that are "behind" that cell and between the two rays are
    invisible to the observing agent. In the mask, 1 means that the cell is visibile,
    0 means that it is invisible.

    Args:
        agent: The agent of interest.
        grid: The grid.
        mask_range: The integer range from the agent of interest.
        agents: The dictionary of agents.

    Returns:
        Two matrices. The first is a local grid centered around the agent's location
        with values from the actual grid. The second is a mask of size
        (2 * range + 1) x (2 * range + 1) that shows which cells are masked to the
        agent of interest.
    """
    # Generate a completely empty grid
    local_grid = np.empty((mask_range * 2 + 1, mask_range * 2 + 1), dtype=object)

    # Copy the section of the grid around the agent's position
    (r, c) = agent.position
    r_lower = max([0, r - mask_range])
    r_upper = min([grid.rows - 1, r + mask_range]) + 1
    c_lower = max([0, c - mask_range])
    c_upper = min([grid.cols - 1, c + mask_range]) + 1
    local_grid[
        (r_lower+mask_range-r):(r_upper+mask_range-r),
        (c_lower+mask_range-c):(c_upper+mask_range-c)
    ] = grid[r_lower:r_upper, c_lower:c_upper]

    mask = np.ones((2 * mask_range + 1, 2 * mask_range + 1))
    for other in agents.values():
        if other.blocking:
            r_diff, c_diff = other.position - agent.position
            # Ensure the other agent within the view range
            if -mask_range <= r_diff <= mask_range and \
                    -mask_range <= c_diff <= mask_range:
                if c_diff > 0 and r_diff == 0: # Other is to the right of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff - 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff - 0.5) * t
                    for c in range(c_diff, mask_range+1):
                        for r in range(-mask_range, mask_range+1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff > 0 and r_diff > 0: # Other is below-right of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff - 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff + 0.5) * t
                    for c in range(c_diff, mask_range+1):
                        for r in range(r_diff, mask_range+1):
                            if c == c_diff and r == r_diff: continue # Don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff == 0 and r_diff > 0: # Other is below the agent
                    left = lambda t: (c_diff - 0.5) / (r_diff - 0.5) * t
                    right = lambda t: (c_diff + 0.5) / (r_diff - 0.5) * t
                    for c in range(-mask_range, mask_range+1):
                        for r in range(r_diff, mask_range+1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if left(r) < c < right(r):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff < 0 and r_diff > 0: # Other is below-left of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff + 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff - 0.5) * t
                    for c in range(c_diff, -mask_range-1, -1):
                        for r in range(r_diff, mask_range+1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff < 0 and r_diff == 0: # Other is left of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff + 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff + 0.5) * t
                    for c in range(c_diff, -mask_range-1, -1):
                        for r in range(-mask_range, mask_range+1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff < 0 and r_diff < 0: # Other is above-left of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff - 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff + 0.5) * t
                    for c in range(c_diff, -mask_range - 1, -1):
                        for r in range(r_diff, -mask_range - 1, -1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff == 0 and r_diff < 0: # Other is above the agent
                    left = lambda t: (c_diff - 0.5) / (r_diff + 0.5) * t
                    right = lambda t: (c_diff + 0.5) / (r_diff + 0.5) * t
                    for c in range(-mask_range, mask_range+1):
                        for r in range(r_diff, -mask_range - 1, -1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if left(r) < c < right(r):
                                mask[r + mask_range, c + mask_range] = 0
                elif c_diff > 0 and r_diff < 0: # Other is above-right of agent
                    upper = lambda t: (r_diff + 0.5) / (c_diff + 0.5) * t
                    lower = lambda t: (r_diff - 0.5) / (c_diff - 0.5) * t
                    for c in range(c_diff, mask_range+1):
                        for r in range(r_diff, -mask_range - 1, -1):
                            if c == c_diff and r == r_diff: continue # don't mask the other
                            if lower(c) < r < upper(c):
                                mask[r + mask_range, c + mask_range] = 0

    return local_grid, mask



import numpy as np


def array_in_array(search_element, elements, axis=0):
    """
    Exact check for a a search element within a collection of elements. Te search element is a
    numpy array, and hte elements is a numpy ndarray. We treat the search element as a single
    entity, so the match has to be across the entire array, not just on the elements within the
    array.

    Args:
        search_element: numpy array that we want to match
        elements: numpy ndarray that we search for a match
        axis: which axis we want to search along. Support 0 and 1.
    """
    if axis == 1:
        elements = elements.transpose()
    for test_element in elements:
        if np.all(search_element == test_element):
            return True
    return False



from abc import ABC, abstractmethod

from ray.rllib.policy.policy import Policy


class HeuristicPolicy(Policy, ABC):
    """Abstract class for policies that do not learn."""
    @abstractmethod
    def compute_actions(self, *args, **kwargs):
        pass

    def learn_on_batch(self, samples):
        pass

    def get_weights(self):
        pass

    def set_weights(self, weights):
        pass


class RandomAction(HeuristicPolicy):
    """A policy to take random actions."""
    def __init__(self, observation_space, action_space, config={}):
        super().__init__(observation_space, action_space, config)

    def compute_actions(self,
                        obs_batch,
                        state_batches=None,
                        prev_action_batch=None,
                        prev_reward_batch=None,
                        info_batch=None,
                        episodes=None,
                        **kwargs):
        return [self.action_space.sample() for _ in obs_batch], [], {}

    def learn_on_batch(self, samples):
        pass

    def get_weights(self):
        pass

    def set_weights(self, weights):
        pass



def run(sim, trainer):
    """
    Analyze the behavior of your trained policies using the simulation and trainer
    from your RL experiment.

    Args:
        sim:
            Simulation Manager object from the experiment.
        trainer:
            Trainer that computes actions using the trained policies.
    """
    # Run the simulation with actions chosen from the trained policies
    policy_agent_mapping = trainer.config['multiagent']['policy_mapping_fn']
    for episode in range(100):
        print('Episode: {}'.format(episode))
        obs = sim.reset()
        done = {agent: False for agent in obs}
        while True: # Run until the episode ends
            # Get actions from policies
            joint_action = {}
            for agent_id, agent_obs in obs.items():
                if done[agent_id]: continue # Don't get actions for done agents
                policy_id = policy_agent_mapping(agent_id)
                action = trainer.compute_action(agent_obs, policy_id=policy_id)
                joint_action[agent_id] = action
            # Step the simulation
            obs, reward, done, info = sim.step(joint_action)
            if done['__all__']:
                break




from gym.spaces import Dict, Discrete

from abmarl.sim.agent_based_simulation import ActingAgent, ObservingAgent, Agent
from abmarl.sim.wrappers import Wrapper
from abmarl.tools import gym_utils as gu


class SuperAgentWrapper(Wrapper):
    """
    The SuperAgentWrapper creates "super" agents who cover and control multiple agents.

    The super agents take the observation and action spaces of all their covered
    agents. In addition, the observation space is given a "mask" channel to indicate
    which of their covered agents is done. This channel is important because
    the simulation dynamics change when a covered agent is done but the super agent
    may still be active (see comments on get_done). Without this mask, the super
    agent would experience completely different simulation dynamcis for some of
    its covered agents with no indication as to why.

    Furthermore, super agents may still report actions for covered agents that
    are done. This wrapper filters out those actions before passing them to the
    underlying sim. See step for more details.
    """
    def __init__(self, sim, super_agent_mapping=None, **kwargs):
        self.sim = sim
        self.super_agent_mapping = super_agent_mapping

    @property
    def super_agent_mapping(self):
        """
        A dictionary that maps from a super agent's id to a list of covered agent ids.

        Suppose our simulation has 5 agents and we use the following super agent mapping:
            {'super0': ['agent0', 'agent1'], 'super1': ['agent3', 'agent4']}
        The resulting agents dict would have keys 'super0', 'super1', and 'agent2';
        where 'agent0', 'agent1', 'agent3', and 'agent4' have been covered by the
        super agents and 'agent2' is left uncovered and therefore included in the
        dict of agents. If the super agent mapping is changed, then the dictionary
        of agents gets recreated immediately.

        Super agents cannot have the same id as any of the agents in the simulation.
        Two super agents cannot cover the same agent. All covered agents must be
        learning agents.
        """
        return self._super_agent_mapping

    @super_agent_mapping.setter
    def super_agent_mapping(self, value):
        assert type(value) is dict, "super agent mapping must be a dictionary."
        self._covered_agents = set()
        for k, v in value.items():
            assert type(k) is str, "The keys super agent mapping must be the super agent's id."
            assert k not in self.sim.agents, \
                "A super agent cannot have the same id as an agent from the underlying sim."
            assert type(v) is list, "The values in super agent mapping must be lists of agent ids."
            for covered_agent in v:
                assert type(covered_agent) is str, "The covered agents list must be agent ids."
                assert covered_agent in self.sim.agents, \
                    "The covered agent must be an agent in the underlying sim."
                assert covered_agent not in self._covered_agents, \
                    "The agent is already covered by another super agent."
                assert isinstance(self.sim.agents[covered_agent], ActingAgent) and \
                    isinstance(self.sim.agents[covered_agent], ObservingAgent), \
                    "Covered agents must be learning Agents."
                self._covered_agents.add(covered_agent)
        self._uncovered_agents = self.sim.agents.keys() - self._covered_agents
        self._super_agent_mapping = value
        # We need to reconstruct the agent dictionary if the super agent mapping
        # ever changes
        self._construct_agents_from_super_agent_mapping()

    def step(self, action_dict, **kwargs):
        """
        Give actions to the simulation.

        Super agent actions are decomposed into the covered agent actions and
        then passed to the underlying sim. Because of the nature of this wrapper,
        the super agents may provide actions for covered agents that are already
        done. We filter out these actions.

        Args:
            action_dict: Dictionary that maps agent ids to the actions. Covered
            agents should not be present.
        """
        unravelled_action_dict = {}
        for agent_id, action in action_dict.items():
            assert agent_id not in self._covered_agents, \
                "We cannot receive actions from an agent that is covered by a super agent."
            if agent_id in self.super_agent_mapping: # A super agent action
                for covered_agent_id, covered_action in action.items():
                    # We can safely assume the format of the actions because we
                    # generated the action space
                    if not self.sim.get_done(covered_agent_id):
                        # We don't want to send the simulation actions from covered
                        # agents that are done
                        unravelled_action_dict[covered_agent_id] = covered_action
            else:
                unravelled_action_dict[agent_id] = action
        self.sim.step(unravelled_action_dict, **kwargs)

    def get_obs(self, agent_id, **kwargs):
        """
        Report observations from the simulation.

        Super agent observations are collected from their covered agents. Super
        agents also have a "mask" channel that tells them which of their covered
        agent is done. This should assist the super agent in understanding the
        changing simulation dynamics for done agents (i.e. why actions from done
        agents don't do anything).

        Args:
            agent_id: The id of the agent for whom to produce an observation. Should
            not be a covered agent.

        Returns:
            The requested observation. Super agent observations are collected from
            the covered agents.
        """
        assert agent_id not in self._covered_agents, \
            "We cannot produce observations for an agent that is covered by a super agent."
        # We can safely assume the format of the observations because we generated
        # the observation space
        if agent_id in self.super_agent_mapping:
            obs = {'mask': {}}
            for covered_agent in self.super_agent_mapping[agent_id]:
                obs[covered_agent] = self.sim.get_obs(covered_agent, **kwargs)
                obs['mask'][covered_agent] = False if self.sim.get_done(covered_agent) else True
            return obs
        else:
            return self.sim.get_obs(agent_id, **kwargs)

    def get_reward(self, agent_id, **kwargs):
        """
        Report the agent's reward.

        Args:
            agent_id: The id of the agent for whom to report the reward. Should
            not be a covered agent.

        Returns:
            The requested reward. Super agent rewards are summed from the covered
            agents.
        """
        assert agent_id not in self._covered_agents, \
            "We cannot get rewards for an agent that is covered by a super agent."
        if agent_id in self.super_agent_mapping:
            return sum([
                self.sim.get_reward(covered_agent_id)
                for covered_agent_id in self.super_agent_mapping[agent_id]
            ])
        else:
            return self.sim.get_reward(agent_id, **kwargs)

    def get_done(self, agent_id, **kwargs):
        """
        Report the agent's done condition.

        Because super agents are composed of multiple agents, it could be the case
        that some covered agents are done while other are not for the same super
        agent. Because we still want those non-done agents to interact with the
        simulation, the super agent only reports done when ALL of its covered agents
        report done.

        Args:
            agent_id: The id of the agent for whom to report the done condition.
            Should not be a covered agent.

        Returns:
            The requested done conndition. Super agents are done when all their
            covered agents are done.
        """
        assert agent_id not in self._covered_agents, \
            "We cannot get done for an agent that is covered by a super agent."
        if agent_id in self.super_agent_mapping:
            return all([
                self.sim.get_done(covered_agent_id)
                for covered_agent_id in self.super_agent_mapping[agent_id]
            ])
        else:
            return self.sim.get_done(agent_id, **kwargs)

    def get_info(self, agent_id, **kwargs):
        """
        Report the agent's additional info.

        Args:
            agent_id: The id of the agent for whom to get info. Should not be a
            covered agent.

        Returns:
            The requested info. Super agents info is collected from covered agents.
        """
        assert agent_id not in self._covered_agents, \
            "We cannot get info for an agent that is covered by a super agent."
        if agent_id in self.super_agent_mapping:
            return {
                covered_agent_id: self.sim.get_info(covered_agent_id, **kwargs)
                for covered_agent_id in self.super_agent_mapping[agent_id]
            }
        else:
            return self.sim.get_info(agent_id, **kwargs)

    def _construct_agents_from_super_agent_mapping(self):
        agents = {}

        # Construct the agent dict with super agents
        for super_agent_id, covered_agent_list in self.super_agent_mapping.items():
            # Construct a mapping from the super agents to the covered agents' observation
            # and action spaces
            obs_mapping = {'mask': {}}
            for covered_agent_id in covered_agent_list:
                obs_mapping[covered_agent_id] = self.sim.agents[covered_agent_id].observation_space
                obs_mapping['mask'][covered_agent_id] = Discrete(2)
            action_mapping = {
                covered_agent_id: self.sim.agents[covered_agent_id].action_space
                for covered_agent_id in covered_agent_list
            }
            agents[super_agent_id] = Agent(
                id=super_agent_id,
                observation_space=gu.make_dict(obs_mapping),
                action_space=Dict(action_mapping)
            )

        # Add all uncovered agents to the dict of agetns
        for agent_id in self._uncovered_agents:
            agents[agent_id] = self.sim.agents[agent_id]

        self.agents = agents



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='abmarl',
    version='0.2.3',
    description='Agent Based Simulation and MultiAgent Reinforcement Learning',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/llnl/abmarl',
    author='Edward Rusu',
    author_email='rusu1@llnl.gov',
    license='BSD 3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    project_urls={
        'Featured Usage': 'https://abmarl.readthedocs.io/en/latest/featured_usage.html',
        'Documentation': 'https://abmarl.readthedocs.io/en/latest/index.html',
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'tensorflow',
        'gym<0.22',
        'ray[rllib]==1.12.1',
        'matplotlib',
        'seaborn',
    ],
    python_requires='>=3.7, <3.9',
    entry_points={
        'console_scripts': [
            'abmarl=abmarl.scripts.scripts:cli'
        ]
    },
)




from gym.spaces import Dict
from ray.rllib import MultiAgentEnv

from abmarl.sim.agent_based_simulation import ActingAgent, ObservingAgent


class MultiAgentWrapper(MultiAgentEnv):
    """
    Enable connection between SimulationManager and RLlib Trainer.

    Wraps a SimulationManager and forwards all calls to the manager. This class
    is boilerplate and needed because RLlib checks that the simulation is an instance
    of MultiAgentEnv.

    Attributes:
        sim: The SimulationManager.
    """
    def __init__(self, sim):
        from abmarl.managers import SimulationManager
        assert isinstance(sim, SimulationManager)
        self.sim = sim

        self._agent_ids = set(
            agent.id for agent in self.sim.agents.values()
            if isinstance(agent, ActingAgent) and isinstance(agent, ObservingAgent)
        )
        self.observation_space = Dict({
            agent.id: agent.observation_space
            for agent in self.sim.agents.values()
            if isinstance(agent, ObservingAgent)
        })
        self.action_space = Dict({
            agent.id: agent.action_space
            for agent in self.sim.agents.values()
            if isinstance(agent, ActingAgent)
        })

    def reset(self):
        """See SimulationManager."""
        return self.sim.reset()

    def step(self, actions):
        """See SimulationManager."""
        return self.sim.step(actions)

    def render(self, *args, **kwargs):
        """See SimulationManager."""
        return self.sim.render(*args, **kwargs)



import os

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import ray
import ray.rllib
from ray.rllib.env import MultiAgentEnv
from ray.tune.registry import get_trainable_cls

from abmarl.tools import utils as adu
from abmarl.managers import SimulationManager


def _start(full_trained_directory, requested_checkpoint, seed=None):
    """The elements that are common to both analyze and visualize."""
    # Load the experiment as a module
    # First, we must find the .py file in the directory
    py_files = [file for file in os.listdir(full_trained_directory) if file.endswith('.py')]
    assert len(py_files) == 1
    full_path_to_config = os.path.join(full_trained_directory, py_files[0])
    experiment_mod = adu.custom_import_module(full_path_to_config)
    # Modify the number of workers in the configuration
    experiment_mod.params['ray_tune']['config']['num_workers'] = 1
    experiment_mod.params['ray_tune']['config']['num_envs_per_worker'] = 1
    experiment_mod.params['ray_tune']['config']['seed'] = seed

    checkpoint_dir, checkpoint_value = adu.checkpoint_from_trained_directory(
        full_trained_directory, requested_checkpoint
    )
    print(checkpoint_dir)

    # Setup ray
    ray.init()

    # Get the trainer
    alg = get_trainable_cls(experiment_mod.params['ray_tune']['run_or_experiment'])
    trainer = alg(
        env=experiment_mod.params['ray_tune']['config']['env'],
        config=experiment_mod.params['ray_tune']['config']
    )
    trainer.restore(os.path.join(checkpoint_dir, 'checkpoint-' + str(checkpoint_value)))

    # Get the simulation
    sim = experiment_mod.params['experiment']['sim_creator'](
        experiment_mod.params['ray_tune']['config']['env_config']
    )

    return sim, trainer


def _finish():
    """Finish off the evaluation run."""
    ray.shutdown()


def run_analysis(full_trained_directory, full_subscript, parameters):
    """Analyze MARL policies from a saved policy through an analysis script"""
    sim, trainer = _start(full_trained_directory, parameters.checkpoint, parameters.seed)

    # The sim may be wrapped by an external wrapper, which we support, but we need
    # to unwrap it.
    if not isinstance(sim, SimulationManager):
        sim = sim.unwrapped

    # Load the analysis module and run it
    analysis_mod = adu.custom_import_module(full_subscript)
    analysis_mod.run(sim, trainer)

    _finish()


def run_visualize(full_trained_directory, parameters):
    """Visualize MARL policies from a saved policy"""
    sim, trainer = _start(full_trained_directory, parameters.checkpoint, seed=parameters.seed)

    # Determine if we are single- or multi-agent case.
    def _multi_get_action(obs, done=None, sim=None, policy_agent_mapping=None, **kwargs):
        joint_action = {}
        if done is None:
            done = {agent: False for agent in obs}
        for agent_id, agent_obs in obs.items():
            if done[agent_id]: continue # Don't get actions for done agents
            policy_id = policy_agent_mapping(agent_id)
            action = trainer.compute_action(
                agent_obs, policy_id=policy_id, explore=parameters.no_explore
            )
            joint_action[agent_id] = action
        return joint_action

    def _single_get_action(obs, trainer=None, **kwargs):
        return trainer.compute_action(obs, explore=parameters.no_explore)

    def _multi_get_done(done):
        return done['__all__']

    def _single_get_done(done):
        return done

    policy_agent_mapping = None
    if isinstance(sim, MultiAgentEnv):
        policy_agent_mapping = trainer.config['multiagent']['policy_mapping_fn']
        _get_action = _multi_get_action
        _get_done = _multi_get_done
    else:
        _get_action = _single_get_action
        _get_done = _single_get_done

    for episode in range(parameters.episodes):
        print('Episode: {}'.format(episode))
        obs = sim.reset()
        done = None
        all_done = False
        fig = plt.figure()

        def gen_frame_until_done():
            nonlocal all_done
            i = 0
            while not all_done:
                i += 1
                yield i

        def animate(i):
            nonlocal obs, done
            sim.render(fig=fig)
            plt.pause(1e-16)
            action = _get_action(
                obs, done=done, sim=sim, trainer=trainer, policy_agent_mapping=policy_agent_mapping
            )
            obs, _, done, _ = sim.step(action)
            if _get_done(done) or i >= parameters.steps_per_episode:
                nonlocal all_done
                all_done = True
                sim.render(fig=fig)
                plt.pause(1e-16)

        anim = FuncAnimation(
            fig, animate, frames=gen_frame_until_done, repeat=False,
            interval=parameters.frame_delay
        )
        if parameters.record:
            anim.save(os.path.join(full_trained_directory, 'Episode_{}.mp4'.format(episode)))
        plt.show(block=False)
        while not all_done:
            plt.pause(1)
        plt.close(fig)

    _finish()



import copy

from gym.spaces import Box, Discrete, Tuple, Dict, MultiDiscrete, MultiBinary
import numpy as np

from .sar_wrapper import SARWrapper


def flatdim(space):
    """Return the number of dimensions a flattened equivalent of this space
    would have.

    Accepts a space and returns an integer. Raises TypeError if
    the space is not defined in gym.spaces.
    """
    if isinstance(space, Box):
        return int(np.prod(space.shape))
    elif isinstance(space, Discrete):
        return int(space.n)
    elif isinstance(space, Tuple):
        return int(sum([flatdim(s) for s in space.spaces]))
    elif isinstance(space, Dict):
        return int(sum([flatdim(s) for s in space.spaces.values()]))
    elif isinstance(space, MultiBinary):
        return int(space.n)
    elif isinstance(space, MultiDiscrete):
        return int(np.prod(space.shape))
    else:
        raise TypeError


def flatten(space, x):
    """Flatten a data point from a space.

    This is useful when e.g. points from spaces must be passed to a neural
    network, which only understands flat arrays of floats.

    Accepts a space and a point from that space. Always returns a 1D array.
    Raises TypeError if the space is not a gym space.
    """
    if isinstance(space, Box):
        return np.asarray(x, dtype=space.dtype).flatten()
    elif isinstance(space, Discrete):
        onehot = np.zeros(space.n, dtype=int)
        onehot[x] = 1
        return onehot
    elif isinstance(space, Tuple):
        return np.concatenate([flatten(s, x_part) for x_part, s in zip(x, space.spaces)])
    elif isinstance(space, Dict):
        return np.concatenate(
            [flatten(s, x[key]) for key, s in space.spaces.items()])
    elif isinstance(space, MultiBinary):
        return np.asarray(x, dtype=int).flatten()
    elif isinstance(space, MultiDiscrete):
        return np.asarray(x, dtype=int).flatten()
    else:
        raise TypeError('space must be instance of gym.spaces')


def unflatten(space, x):
    """Unflatten a data point from a space.

    This reverses the transformation applied by flatten(). You must ensure
    that the space argument is the same as for the flatten() call.

    Accepts a space and a flattened point. Returns a point with a structure
    that matches the space. Raises TypeError if the space is not
    defined in gym.spaces.
    """
    if isinstance(space, Box):
        return np.asarray(x, dtype=space.dtype).reshape(space.shape)
    elif isinstance(space, Discrete):
        return int(np.nonzero(x)[0][0])
    elif isinstance(space, Tuple):
        dims = [flatdim(s) for s in space.spaces]
        list_flattened = np.split(x, np.cumsum(dims)[:-1])
        list_unflattened = [
            unflatten(s, flattened)
            for flattened, s in zip(list_flattened, space.spaces)
        ]
        return tuple(list_unflattened)
    elif isinstance(space, Dict):
        dims = [flatdim(s) for s in space.spaces.values()]
        list_flattened = np.split(x, np.cumsum(dims)[:-1])
        list_unflattened = [
            (key, unflatten(s, flattened))
            for flattened, (key,
                            s) in zip(list_flattened, space.spaces.items())
        ]
        from collections import OrderedDict
        return OrderedDict(list_unflattened)
    elif isinstance(space, MultiBinary):
        return np.asarray(x, dtype=int).reshape(space.shape)
    elif isinstance(space, MultiDiscrete):
        return np.asarray(x, dtype=int).reshape(space.shape)
    else:
        raise TypeError


def flatten_space(space):
    """Flatten a space into a single Box.

    This is equivalent to flatten(), but operates on the space itself. The
    result always is a Box with flat boundaries. The box has exactly
    flatdim(space) dimensions. Flattening a sample of the original space
    has the same effect as taking a sample of the flattenend space.

    Raises TypeError if the space is not defined in gym.spaces.

    Example::

        >>> box = Box(0.0, 1.0, shape=(3, 4, 5))
        >>> box
        Box(3, 4, 5)
        >>> flatten_space(box)
        Box(60,)
        >>> flatten(box, box.sample()) in flatten_space(box)
        True

    Example that flattens a discrete space::

        >>> discrete = Discrete(5)
        >>> flatten_space(discrete)
        Box(5,)
        >>> flatten(box, box.sample()) in flatten_space(box)
        True

    Example that recursively flattens a dict::

        >>> space = Dict({"position": Discrete(2),
        ...               "velocity": Box(0, 1, shape=(2, 2))})
        >>> flatten_space(space)
        Box(6,)
        >>> flatten(space, space.sample()) in flatten_space(space)
        True
    """
    if isinstance(space, Box):
        return Box(space.low.flatten(), space.high.flatten(), dtype=space.dtype)
    if isinstance(space, Discrete):
        return Box(low=0, high=1, shape=(space.n, ), dtype=int)
    if isinstance(space, Tuple):
        space = [flatten_space(s) for s in space.spaces]
        encapsulating_type = int \
            if all([this_space.dtype == int for this_space in space]) \
            else float
        return Box(
            low=np.concatenate([s.low for s in space]),
            high=np.concatenate([s.high for s in space]),
            dtype=encapsulating_type
        )
    if isinstance(space, Dict):
        space = [flatten_space(s) for s in space.spaces.values()]
        encapsulating_type = int \
            if all([this_space.dtype == int for this_space in space]) \
            else float
        return Box(
            low=np.concatenate([s.low for s in space]),
            high=np.concatenate([s.high for s in space]),
            dtype=encapsulating_type
        )
    if isinstance(space, MultiBinary):
        return Box(low=0, high=1, shape=(space.n, ), dtype=int)
    if isinstance(space, MultiDiscrete):
        return Box(
            low=np.zeros_like(space.nvec),
            high=space.nvec,
            dtype=int
        )
    raise TypeError


class FlattenWrapper(SARWrapper):
    """
    Flattens all agents' action and observation spaces into continuous Boxes.
    """
    def __init__(self, sim):
        super().__init__(sim)
        for agent_id, wrapped_agent in self.sim.agents.items(): # Wrap the agents' spaces
            self.agents[agent_id].action_space = flatten_space(wrapped_agent.action_space)
            self.agents[agent_id].observation_space = flatten_space(
                wrapped_agent.observation_space
            )

    def wrap_observation(self, from_agent, observation):
        return flatten(from_agent.observation_space, observation)

    def unwrap_observation(self, from_agent, observation):
        return unflatten(from_agent.observation_space, observation)

    def wrap_action(self, from_agent, action):
        return unflatten(from_agent.action_space, action)

    def unwrap_action(self, from_agent, action):
        return flatten(from_agent.action_space, action)


class FlattenActionWrapper(SARWrapper):
    """
    Flattens all agents' action spaces into continuous Boxes.
    """
    def __init__(self, sim):
        super().__init__(sim)
        self.agents = copy.deepcopy(self.sim.agents)
        for agent_id, wrapped_agent in self.sim.agents.items():
            # Wrap the action spaces of the agents
            self.agents[agent_id].action_space = flatten_space(wrapped_agent.action_space)

    def wrap_action(self, from_agent, action):
        return unflatten(from_agent.action_space, action)

    def unwrap_action(self, from_agent, action):
        return flatten(from_agent.action_space, action)




from gym.spaces import Discrete, Dict, Box
import numpy as np
from matplotlib import pyplot as plt

from abmarl.sim import Agent
from abmarl.sim.gridworld.agent import MovingAgent, GridObservingAgent, GridWorldAgent
from abmarl.sim.gridworld.base import GridWorldSimulation
from abmarl.sim.gridworld.state import PositionState, StateBaseComponent
from abmarl.sim.gridworld.actor import MoveActor, ActorBaseComponent
from abmarl.sim.gridworld.observer import SingleGridObserver, ObserverBaseComponent
from abmarl.sim.gridworld.done import DoneBaseComponent
import abmarl.sim.gridworld.utils as gu


class BroadcastingAgent(Agent, GridWorldAgent):
    def __init__(self, broadcast_range=None, initial_message=None, **kwargs):
        super().__init__(**kwargs)
        self.broadcast_range = broadcast_range
        self.initial_message = initial_message

    @property
    def broadcast_range(self):
        return self._broadcast_range

    @broadcast_range.setter
    def broadcast_range(self, value):
        assert type(value) is int and value >= 0, "Broadcast Range must be a nonnegative integer."
        self._broadcast_range = value

    @property
    def initial_message(self):
        return self._initial_message

    @initial_message.setter
    def initial_message(self, value):
        if value is not None:
            assert -1 <= value <= 1, "Initial message must be a number between -1 and 1."
        self._initial_message = value

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, value):
        self._message = min(max(value, -1), 1)

    @property
    def configured(self):
        return super().configured and self.broadcast_range is not None


class BroadcastingActor(ActorBaseComponent):
    """
    Process sending and receiving messages between agents.

    Broadcasting Agents can broadcast to agents within their range according to
    the broadcast mapping and if the agent is not blocked.
    """
    def __init__(self, broadcast_mapping=None, **kwargs):
        super().__init__(**kwargs)
        self.broadcast_mapping = broadcast_mapping
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.action_space[self.key] = Discrete(2)

    @property
    def key(self):
        return 'broadcast'

    @property
    def supported_agent_type(self):
        return BroadcastingAgent

    @property
    def broadcast_mapping(self):
        """
        Dict that dictates to which agents the broadcasting agent can broadcast.

        The dictionary maps the broadcasting agents' encodings to a list of encodings
        to which they can broadcast. For example, the folowing broadcast_mapping:
        {
            1: [3, 4, 5],
            3: [2, 3],
        }
        means that agents whose encoding is 1 can broadcast other agents whose encodings
        are 3, 4, or 5; and agents whose encoding is 3 can broadcast other agents whose
        encodings are 2 or 3.
        """
        return self._broadcast_mapping

    @broadcast_mapping.setter
    def broadcast_mapping(self, value):
        assert type(value) is dict, "Broadcast mapping must be dictionary."
        for k, v in value.items():
            assert type(k) is int, "All keys in broadcast mapping must be integer."
            assert type(v) is list, "All values in broadcast mapping must be list."
            for i in v:
                assert type(i) is int, \
                    "All elements in the broadcast mapping values must be integers."
        self._broadcast_mapping = value

    def process_action(self, broadcasting_agent, action_dict, **kwargs):
        """
        If the agent has chosen to broadcast, then we process their broadcast.

        The processing goes through a series of checks. The broadcast is successful
        if there is a receiving agent such that:
        1. The receiving agent is within range.
        2. The receiving agent is valid according to the broadcast_mapping.
        3. The receiving agent is observable by the broadcasting agent.

        If the broadcast is successful, then the receiving agent receives the message
        in its observation.
        """
        def determine_broadcast(agent):
            # Generate local grid and a broadcast mask.
            local_grid, mask = gu.create_grid_and_mask(
                agent, self.grid, agent.broadcast_range, self.agents
            )

            # Randomly scan the local grid for receiving agents.
            receiving_agents = []
            for r in range(2 * agent.broadcast_range + 1):
                for c in range(2 * agent.broadcast_range + 1):
                    if mask[r, c]: # We can see this cell
                        candidate_agents = local_grid[r, c]
                        if candidate_agents is not None:
                            for other in candidate_agents.values():
                                if other.id == agent.id: # Cannot broadcast to yourself
                                    continue
                                elif other.encoding not in self.broadcast_mapping[agent.encoding]:
                                    # Cannot broadcast to this type of agent
                                    continue
                                else:
                                    receiving_agents.append(other)
            return receiving_agents

        if isinstance(broadcasting_agent, self.supported_agent_type):
            action = action_dict[self.key]
            if action: # Agent has chosen to attack
                return determine_broadcast(broadcasting_agent)


class BroadcastingState(StateBaseComponent):
    def reset(self, **kwargs):
        for agent in self.agents.values():
            if isinstance(agent, BroadcastingAgent):
                if agent.initial_message is not None:
                    agent.message = agent.initial_message
                else:
                    agent.message = np.random.uniform(-1, 1)

        # Tracks agents receiving messages from other agents
        self.receiving_state = {
            agent.id: [] for agent in self.agents.values() if isinstance(agent, BroadcastingAgent)
        }

    def update_receipients(self, from_agent, to_agents):
        for agent in to_agents:
            self.receiving_state[agent.id].append((from_agent.id, from_agent.message))

    def update_message_and_reset_receiving(self, agent):
        receiving_from = self.receiving_state[agent.id]
        self.receiving_state[agent.id] = []

        messages = [message for _, message in receiving_from]
        messages.append(agent.message)
        agent.message = np.average(messages)

        return receiving_from


class BroadcastObserver(ObserverBaseComponent):
    def __init__(self, broadcasting_state=None, **kwargs):
        super().__init__(**kwargs)

        assert isinstance(broadcasting_state, BroadcastingState), \
            "broadcasting_state must be an instance of BroadcastingState"
        self._broadcasting_state = broadcasting_state

        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.observation_space[self.key] = Dict({
                    other.id: Box(-1, 1, (1,))
                    for other in self.agents.values()
                    if isinstance(other, self.supported_agent_type)
                })

    @property
    def key(self):
        return 'message'

    @property
    def supported_agent_type(self):
        return BroadcastingAgent

    def get_obs(self, agent, **kwargs):
        if not isinstance(agent, self.supported_agent_type):
            return {}

        obs = {other: 0 for other in agent.observation_space[self.key]}
        receive_from = self._broadcasting_state.update_message_and_reset_receiving(agent)
        for agent_id, message in receive_from:
            obs[agent_id] = message
        obs[agent.id] = agent.message
        return obs


class AverageMessageDone(DoneBaseComponent):
    def __init__(self, done_tolerance=None, **kwargs):
        super().__init__(**kwargs)
        self.done_tolerance = done_tolerance

    @property
    def done_tolerance(self):
        return self._done_tolerance

    @done_tolerance.setter
    def done_tolerance(self, value):
        assert type(value) in [int, float], "Done tolerance must be a number."
        assert value > 0, "Done tolerance must be positive."
        self._done_tolerance = value

    def get_done(self, agent, **kwargs):
        if isinstance(agent, BroadcastingAgent):
            average = np.average([
                other.message for other in self.agents.values()
                if isinstance(other, BroadcastingAgent)
            ])
            return np.abs(agent.message - average) <= self.done_tolerance
        else:
            return False

    def get_all_done(self, **kwargs):
        for agent in self.agents.values():
            if isinstance(agent, BroadcastingAgent):
                if not self.get_done(agent):
                    return False
        return True


class BlockingAgent(MovingAgent, GridObservingAgent):
    def __init__(self, **kwargs):
        super().__init__(blocking=True, **kwargs)


class BroadcastSim(GridWorldSimulation):
    def __init__(self, **kwargs):
        self.agents = kwargs['agents']

        self.position_state = PositionState(**kwargs)
        self.broadcasting_state = BroadcastingState(**kwargs)

        self.move_actor = MoveActor(**kwargs)
        self.broadcast_actor = BroadcastingActor(**kwargs)

        self.grid_observer = SingleGridObserver(**kwargs)
        self.broadcast_observer = BroadcastObserver(
            broadcasting_state=self.broadcasting_state, **kwargs
        )

        self.done = AverageMessageDone(**kwargs)

        self.finalize()

    def reset(self, **kwargs):
        self.position_state.reset(**kwargs)
        self.broadcasting_state.reset(**kwargs)

        self.rewards = {agent.id: 0 for agent in self.agents.values()}

    def step(self, action_dict, **kwargs):
        # process broadcasts
        for agent_id, action in action_dict.items():
            agent = self.agents[agent_id]
            receiving_agents = self.broadcast_actor.process_action(agent, action, **kwargs)
            if receiving_agents is not None:
                self.broadcasting_state.update_receipients(agent, receiving_agents)

        # process moves
        for agent_id, action in action_dict.items():
            agent = self.agents[agent_id]
            move_result = self.move_actor.process_action(agent, action, **kwargs)
            if not move_result:
                self.rewards[agent.id] -= 0.1

        # Entropy penalty
        for agent_id in action_dict:
            self.rewards[agent_id] -= 0.01

    def render(self, **kwargs):
        super().render(**kwargs)
        for agent in self.agents.values():
            if isinstance(agent, BroadcastingAgent):
                print(f"{agent.id}: {agent.message}")
        print()

    def get_obs(self, agent_id, **kwargs):
        agent = self.agents[agent_id]
        return {
            **self.grid_observer.get_obs(agent, **kwargs),
            **self.broadcast_observer.get_obs(agent, **kwargs)
        }

    def get_reward(self, agent_id, **kwargs):
        reward = self.rewards[agent_id]
        self.rewards[agent_id] = 0
        return reward

    def get_done(self, agent_id, **kwargs):
        return self.done.get_done(agent_id, **kwargs)

    def get_all_done(self, **kwargs):
        return self.done.get_all_done(**kwargs)

    def get_info(self, **kwargs):
        return {}


if __name__ == "__main__":
    agents = {
        'broadcaster0': BroadcastingAgent(
            id='broadcaster0', encoding=1, broadcast_range=6, render_color='green'
        ),
        'broadcaster1': BroadcastingAgent(
            id='broadcaster1', encoding=1, broadcast_range=6, render_color='green'
        ),
        'broadcaster2': BroadcastingAgent(
            id='broadcaster2', encoding=1, broadcast_range=6, render_color='green'
        ),
        'broadcaster3': BroadcastingAgent(
            id='broadcaster3', encoding=1, broadcast_range=6, render_color='green'
        ),
        'blocker0': BlockingAgent(
            id='blocker0', encoding=2, move_range=2, view_range=3, render_color='black'
        ),
        'blocker1': BlockingAgent(
            id='blocker1', encoding=2, move_range=1, view_range=3, render_color='black'
        ),
        'blocker2': BlockingAgent(
            id='blocker2', encoding=2, move_range=1, view_range=3, render_color='black'
        ),
    }
    sim = BroadcastSim.build_sim(
        7, 7, agents=agents, broadcast_mapping={1: [1]}, done_tolerance=5e-10
    )

    sim.reset()
    fig = plt.figure()
    sim.render(fig=fig)

    done_agents = set()
    for i in range(50):
        action = {
            agent.id: agent.action_space.sample()
            for agent in agents.values() if agent.id not in done_agents
        }
        sim.step(action)
        for agent in agents:
            if agent not in done_agents:
                obs = sim.get_obs(agent)
            if sim.get_done(agent):
                done_agents.add(agent)

        sim.render(fig=fig)
        if sim.get_all_done():
            break
